;; Generated package description from vertico.el  -*- no-byte-compile: t -*-
(define-package "vertico" "0.29" "VERTical Interactive COmpletion" '((emacs "27.1")) :commit "cacb0bd76153790ba6ecc67921deed236c96371e" :authors '(("Daniel Mendler" . "mail@daniel-mendler.de")) :maintainer '("Daniel Mendler" . "mail@daniel-mendler.de") :url "https://github.com/minad/vertico")
